#include <stdio.h>
#include <sys/socket.h>
#include <netdb.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/poll.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>

#include "common.h"
#include "client.h"

extern "C" int p_poll (struct pollfd*, int, int);


// The infamous

static char *StrDup (char *c)
{
	return strcpy (new char [strlen (c) + 1], c);
}

static int Init__Init = initLOCMessages ();
/**************************** NETWORKING UTILITIES **************************/

static unsigned int nslookup (char *h)
{
	struct hostent *ho;

	if (!(ho = gethostbyname (h)))
		return 0;

	return *(unsigned int*)*ho->h_addr_list;
}

static int connect_tcp (unsigned int IP, short port)
{
	struct sockaddr_in server;
	int Socket;

	if ((Socket = socket (AF_INET, SOCK_STREAM, 0)) == -1)
	{
		perror ("Can't create socket");
		exit (1);
	}

	server.sin_family = AF_INET;
	server.sin_port = htons (port);
	server.sin_addr.s_addr = IP;

	if (connect (Socket, (struct sockaddr*)&server,
	             sizeof (struct sockaddr)) == -1)
	{
		return -1;
	}

	return Socket;
}

static in_addr myIP (int fd)
{
	struct sockaddr_in s;
	socklen_t j = sizeof (s);
	getsockname (fd, (struct sockaddr*)&s, &j);
	return s.sin_addr;
}

/***************************** LOC UTILITIES ****************************/

#define SERVER_TIMEOUT 40*1000

static int getmessage (char *ret, int fd)
{
	struct pollfd pfd  = { fd, POLLIN, 0 };
	int i;

	i = p_poll (&pfd, 1, SERVER_TIMEOUT);
	if (i != 1) return 0;

	i = read (fd, ret, MAX_MSG-1);
	if (i == MAX_MSG-1) i = 0;

	return i;
}

static bool greet_LOC (int fd)
{
	int i;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);
	return i == 2 && MSGtype (argv [0]) == VERSION;
}

/*
 Return a file descriptor to a connection with a LOC user.
 Returns:
	-1 : User not on line
	-2 : Fail lookup for LOC server
	-3 : LOC server not responding
	-4 : Other error
*/
int connect_LOC_user (char *u, char *host, short port)
{
	unsigned int IP;
	int fd, i;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	if ((IP = nslookup (host)) == 0) return -2;

	if ((fd = connect_tcp (IP, port)) == -1) return -3;
	if (!greet_LOC (fd)) return -3;

	argv [0] = u;
	i = mkLOCmessage (buffer, WANNATALKTO, 1, argv);
	write (fd, buffer, i);

	i = getmessage (buffer, fd);
	close (fd);
	i = parseLOCmessage (buffer, i, argv);

	if (i == 0) return -3;

	switch (MSGtype (argv [0])) {
	case NOTCONNECTED:
		return -1;
	case GOTOADDRESS:
		if (i != 3) return -4;
		return connect_tcp (inet_addr (argv [1]), atoi (argv [2]));
	case SETUPRENDEZVOUS:
		if (i != 3) return -4;
		if ((fd = connect_tcp (IP, atoi (argv [1]))) == -1)
			return -4;
		write (fd, argv [2], strlen (argv [2]));
		return fd;
	default:
		break;
	}

	return -4;
}

/*
 negotiate authentication to get into user account mode
*/
static bool LOC_account (char *u, char *p, int fd)
{
	int i;
	md5string M;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	if (!greet_LOC (fd))
		return false;

	argv [0] = u;
	i = mkLOCmessage (buffer, IAM, 1, argv);
	write (fd, buffer, i);

	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);

	if (i != 2 && MSGtype (argv [0]) != AUTHSTR)
		return false;

	argv [0] = do_md5digest (p, argv [1], M);
	i = mkLOCmessage (buffer, DIGEST, 1, argv);
	write (fd, buffer, i);

	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);

	return i == 1 && MSGtype (argv [0]) == OKEY;
}

char *LOCreachable::p;

LOCreachable::LOCreachable (char *us, char *pa, char *ho, short rp, short lp)
{
	int i, fd;
	char *argv [MAX_ARGV], ps [10], buffer [MAX_MSG];

	u = us;
	p = pa;
	port = rp;

	if ((IP = nslookup (ho)) == 0 || (fd = connect_tcp (IP, rp)) == -1) {
		result = -1;
		return;
	}

	if (!LOC_account (u, p, fd)) {
		close (fd);
		result = -2;
		return;
	}

	argv [0] = inet_ntoa (myIP (fd));
	sprintf (ps, "%u", lp);
	argv [1] = ps;
	i = mkLOCmessage (buffer, USERADDRESS, 2, argv);
	write (fd, buffer, i);

	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);

	close (fd);

	result = i == 1 && MSGtype (argv [0]) == OKEY;
}
	
LOCreachable::~LOCreachable ()
{
	int i, fd;
	char buffer [MAX_MSG];

	if ((fd = connect_tcp (IP, port)) == -1)
		 return;

	if (LOC_account (u, p, fd)) {
		i = mkLOCmessage (buffer, DISCONNECT, 0, NULL);
		write (fd, buffer, i);
	}

	close (fd);
}

LOCfirewalled::LOCfirewalled (char *u, char *pa, char *ho, short p)
{
	int i;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	if ((IP = nslookup (ho)) == 0 || (fd = connect_tcp (IP, p)) == -1) {
		result = -1;
		return;
	}

	if (!LOC_account (u, pa, fd)) {
		close (fd);
		result = -2;
		return;
	}

	i = mkLOCmessage (buffer, FIREWALLED, 0, NULL);
	write (fd, buffer, i);

	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);

	if (i != 1 && MSGtype (argv [0]) != OKEY) {
		close (fd);
		result = 0;
	} else result = 1;
}

LOCfirewalled::~LOCfirewalled ()
{
	if (result == 1) close (fd);
}

int LOCfirewalled::accept ()
{
	struct pollfd pfd;
	int i;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	pfd.fd = fd;
	pfd.events = POLLIN;
	pfd.revents = 0;

	i = p_poll (&pfd, 1, -1);
	if (i != 1 || (pfd.revents & POLLHUP)) return -1;
	i = getmessage (buffer, fd);

	i = parseLOCmessage (buffer, i, argv);
	if (i != 3 || MSGtype (argv [0]) != SETUPRENDEZVOUS) return -1;

	int rfd = connect_tcp (IP, atoi (argv [1]));
	if (rfd != -1)
		write (rfd, argv [2], strlen (argv [2]));

	return rfd;
}

/*
*/

bool change_LOC_passwd (char *s, short p, char *u, char *po, char *pn)
{
	int i, fd;
	unsigned int IP;
	bool r;
	char *argv [MAX_ARGV], buffer [MAX_MSG];

	if ((IP = nslookup (s)) == 0) return false;
	if ((fd = connect_tcp (IP, p)) == -1) return false;

	if (!LOC_account (u, po, fd)) {
		close (fd);
		return false;
	}

	argv [0] = pn;
	i = mkLOCmessage (buffer, CHPASS, 1, argv);
	write (fd, buffer, i);
	i = getmessage (buffer, fd);
	i = parseLOCmessage (buffer, i, argv);
	r = i == 1 && MSGtype (argv [0]) == OKEY;

	close (fd);

	// irresponsible but changing your password *is* rare afterall
	LOCreachable::p = StrDup (pn);

	return r;
}

