#include "include.h"

//**************************************
//
//	Post ID -- Filenames
//
//**************************************

bool isPRLaddr (char *s)
{
	return !strncasecmp (s, "/dns/", 5) || !strncasecmp (s, "/loc/", 5);
}

int send_posttext (unsigned int id, int fd)
{
	char filename [10 + sizeof POST_TEXT_FILE];
	sprintf (filename, "%u" POST_TEXT_FILE, id);
	return sendfile (filename, fd);
}

static int access_posttext (unsigned int id, int mode)
{
	char filename [10 + sizeof POST_TEXT_FILE];
	sprintf (filename, "%u" POST_TEXT_FILE, id);
	return access (filename, mode);
}

int rename_ipi (char *f, unsigned int ID)
{
	char filename [10 + sizeof POST_DATA_FILE];
	sprintf (filename, "%u" POST_DATA_FILE, ID);
	return rename (f, filename);
}

int open_postpid (unsigned int id, int flags)
{
	char filename [10 + sizeof POST_DATA_FILE];
	sprintf (filename, "%u" POST_DATA_FILE, id);
	return open (filename, flags);
}

FILE *fopen_postpid (unsigned int id, char *mode)
{
	char filename [10 + sizeof POST_DATA_FILE];
	sprintf (filename, "%u" POST_DATA_FILE, id);
	return fopen (filename, mode);
}

bool parseID (char *t, unsigned int *k, unsigned int *r)
{
	*k = strtoul (t, &t, 16);
	if (*t != '.') return false;
	*r = strtoul (++t, NULL, 16);
	return true;
}

int posttextsize (unsigned int id)
{
	char filename [10 + sizeof POST_TEXT_FILE];
	sprintf (filename, "%u" POST_TEXT_FILE, id);
	return filesize (filename);
}

unsigned long postcrc (unsigned int id)
{
	char filename [10 + sizeof POST_TEXT_FILE];
	sprintf (filename, "%u" POST_TEXT_FILE, id);
	return adler32_file (filename);
}

//****************************************************************
//
// at initialization contact point change detected:
//	- posted posts need update
//	- if reply to self the update is done now
//
// <initialization> = after post info loaded, before server started
//
//****************************************************************

static void internalUpdate (char **ref)
{
	char newref [MAX_PRL];
	sprintf (newref, "%s/%s", Globals.ThisPoint, MsgIDofPRL (*ref));
	delete [] *ref;
	*ref = StrDup (newref);
}

class USetupPost : public intVTask
{
   public:
	void Do (intNode*);
};

void USetupPost::Do (intNode *d)
{
	post *p = (post*) d;
	partial_post *pp = (partial_post*) p->replies.Start;

	if (p->secret != SECRET_UNPOSTED)
		if (p->secret == SECRET_WSELF)
			internalUpdate (&p->parent);
		else p->zombie = true;

	for (; pp; pp = (partial_post*) p->replies.Next (pp))
		if (pp->secret == SECRET_WSELF)
			internalUpdate (&pp->point);
		else pp->zombie = true;
}

void myPointChange ()
{
	USetupPost A;
	Globals.Posts.foreach (&A);
}

//****************************************************************
//
// Queuer state. Go through all posts, see which need
// update or are unposted. Fill queuer entries.
//
//****************************************************************

class QPostInfo : public intVTask
{
   public:
	void Do (intNode*);
};

void QPostInfo::Do (intNode *d)
{
	char recipient [MAX_PRL];
	post *p = (post*) d;
	partial_post *pp = (partial_post*) p->replies.Start;

	if (p->secret == SECRET_UNPOSTED)
		QNewReply (CPofPRL (p->parent, recipient), p->Key);
	else if (p->zombie)
		QChPoint (CPofPRL (p->parent, recipient), p->Key, 0);

	for (; pp; pp = (partial_post*) p->replies.Next (pp))
		if (pp->zombie)
			QChPoint (CPofPRL (pp->point, recipient), p->Key,
				  pp->secret);
}

void QinitDB ()
{
	QPostInfo A;
	Globals.Posts.foreach (&A);
}

//****************************************************************
//
// Load partial post information -- construct database
//	current working directory is the directory of .ipi files
//
//****************************************************************
#define ZEROPOST_TITLE "info"

static unsigned int is_ipi (char *f)
{
	if (strlen (f) >= sizeof POST_DATA_FILE
	&& !strcmp (POST_DATA_FILE, f + strlen (f) - (sizeof POST_DATA_FILE - 1)))
		return strtoul (f, NULL, 10);
	return ~0;
}

struct happy
{
	happy () {
		tx = (getenv ("TERMX")) ? atoi (getenv ("TERMX")) : 80;
		cr = 0;
	} void printi (unsigned int i) {
		char foo [22];
		i = sprintf (foo, " [%u] ", i);
		if (cr + i >= tx) foo [cr = 0] = '\n';
		else cr += i;
		write (fileno (stdout), foo, i);
	} unsigned int cr, tx;
};

void LoadIPI ()
{
	DIR *dp;
	dirent *ep;
	unsigned int ID;
	happy H;

	if (!(dp = opendir ("./"))) exit (1);

	while ((ep = readdir (dp)))
		if ((ID = is_ipi (ep->d_name)) != ~0U) {
			if (Globals.TopPostID <= ID)
				Globals.TopPostID = ID + 1;
			if ((access_posttext (ID, R_OK) == 0 || ID == 0)
			&& !Globals.Posts.intFind (ID)) {
				new post (ID, fopen_postpid (ID, "r"));
				H.printi (ID);
			}
		}

	closedir (dp);

	if (!Globals.Posts.intFind (0)) {
		post *p = new post (0, ZEROPOST_TITLE, NULL);
		p->dump_ipi ();
	}

	printf ("\n");
}

class dump_post : public intVTask
{
   public:
	void Do (intNode*);
};

void dump_post::Do (intNode *n)
{
	post *p = (post*) n;
	p->dump_ipi ();
}

void SaveIPI ()
{
	dump_post A;
	Globals.Posts.foreach (&A);
}

/* HCC RESOLVE:
intTree::walk_tree (P7intNodePFP7intNode_v)
	dump_post (P7intNode)
END */
//***************************************
//	misc/parsers
//***************************************

bool selfref (char *prl)
{
	unsigned int l = strlen (Globals.ThisPoint);
	if (strlen (prl) < l) return false;
	return strncasecmp (prl, Globals.ThisPoint, l) == 0;
}

int slashcount (char *p)
{
	int c = 0;
	while (*p)
		if (*p++ == '/') c++;
	return c;
}

bool goodPRL (char *p)
{
	char *c;
DPRINTF (DB_SLOT4, "GoodPRL(%s)\n", p);
	if (*p != '/' || !(c = strrchr (p + 1, '/'))) return false;
	if (!strncasecmp (p, "/loc/", 5) && slashcount (p) != 5)
		return false;
	else if (!strncasecmp (p, "/dns/", 5) && slashcount (p) != 4)
		return false;
	strtoul (c + 1, &c, 16);
	if (*c != '.') return false;
	strtoul (c + 1, &c, 16);
	return *c == 0;
}

char *MsgIDofPRL (char *prl)
{
	return strrchr (prl, '/') + 1;
}

char *PRLofMsgID (unsigned int k, unsigned int rk, char *ret)
{
	char tmp [20];
	sprintf (tmp, "/%x.%x", k, rk);
	return strcat (strcpy (ret, Globals.ThisPoint), tmp);
}

char *CPofPRL (char *prl, char *ret)
{
	char *c;
	if ((c = strrchr (prl, '/'))) {
		strncpy (ret, prl, c - prl);
		ret [c - prl] = 0;
		return ret;
	}
assert(0);
	return NULL;
}
